import json
from urllib.parse import quote_plus
from urllib.parse import urlparse
import requests
import transaction
import logging

from AccessControl.unauthorized import Unauthorized
from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from plone import api
from plone.rest import Service

from Products.Poi.adapters import IResponseContainer
from Products.Poi.adapters import Response


logger = logging.getLogger('retrans.poi.gitlab')

detail_format = u'''Original POI Issue: {}

### Details

{}

### Steps to Reproduce

{}
'''

class CheckGitlabView(BrowserView):
    """ check for token to display button """

    def show_create_gitlab_issue(self):
        issue = aq_inner(self.context)
        user = api.user.get_current()
        api_token = user.getProperty('gitlab_api_token', None)
        gitlab_username = user.getProperty('gitlab_username', None)
        tracker = issue.getTracker()
        if issue.gitlab_issue_id or not tracker.repo_url or not api_token or not gitlab_username:
            return False
        return True


class CreateGitlabIssueView(BrowserView):
    """ helper for creating gitlab issues """

    def __call__(self):
        response = self.request.response
        issue = aq_inner(self.context)

        user = api.user.get_current()
        api_token = user.getProperty('gitlab_api_token', None)
        tracker = issue.getTracker()
        if not tracker.repo_url or not api_token:
            return 'Please configure a Tracker URL and your GitLab API token.'
        url = urlparse(tracker.repo_url)
        api_root = '{}://{}/api/v4'.format(url.scheme, url.netloc)
        project_id = quote_plus(url.path.strip('/'))
        gitlab_userid = self.get_gitlab_userid(
                api_token,
                user.getProperty('username', None)
            )
        if gitlab_userid is None:
            reason = "Unable to get user's id to assign GitLab issue."
            response.setStatus('500', reason)
            return None
        issue_url = '{}/projects/{}/issues'.format(api_root, project_id)
        issue_detail = detail_format.format(
            issue.absolute_url(),
            issue.details.output,
            issue.steps.output if issue.steps else 'None provided',
        )
        try:
            api_response = requests.post(
                issue_url,
                params={
                    'private_token': api_token,
                },
                data={
                    'id': project_id,
                    'title': '#{} - {}'.format(issue.id, issue.title),
                    'description': issue_detail,
                    'assignee_id': gitlab_userid,
                },
            )
            if api_response.ok:
                gitlab_issue = api_response.json()
                new_issue_url = gitlab_issue.get('web_url', '')
                response = Response(
                    '<a href="{0}">Related GitLab Issue</a>'.format(new_issue_url)
                )
                folder = IResponseContainer(issue)
                folder.add(response)
                issue.gitlab_issue_id = gitlab_issue.get('iid', '')
                transaction.commit()
            else:
                reason = 'Unable to create GitLab issue: {}'.format(api_response.text)
                response.setStatus(api_response.status_code, reason)
                return None
        except Exception as e:
            reason = 'Unable to create GitLab issue: {}'.format(str(e))
            response.setStatus('500', reason)
            return None


    def get_gitlab_userid(self, api_token, username):
        user_url = "http://git.corp.retrans.com/api/v4/users"
        try:
            api_response = requests.get(
                user_url,
                headers={'PRIVATE-TOKEN': api_token,},
                params={
                    'username': username,
                },
            )
            if api_response.ok:
                users = api_response.json()
                if (len(users) > 0):
                    return users[0].get('id', None)
                return None
        except Exception as e:
            return None


class WebhookService(Service):
    """ helper for updating poi from gitlab """

    def render(self):
        request = self.request
        req_token = request.getHeader('X-GitLab-Token')
        secret_token = api.portal.get_registry_record(
            'retrans.poi.webhook_secret_token')
        if req_token != secret_token:
            reason = "Webhook token mismatch."
            logger.error(reason)
            self.request.response.setBody({"error": reason})
            self.request.response.setStatus(500)
            raise Unauthorized(reason)
            return
        event = request.getHeader('X-GitLab-Event')
        data = json.loads(request.get('BODY'))
        username = data.get('user').get('username')

        # breaking out in case we need to handle different
        # kinds of hooks in the future
        if event == 'Issue Hook':
            self.handle_issue_event(username, data)
        return


    def handle_issue_event(self, username, data):
        """ process Issue Hook requests """

        with api.env.adopt_user(username=username):
            # in order to get the correct issue, we need to know which tracker
            # find the tracker based on gitlab project
            project = data.get('project')
            web_url = project.get('web_url')
            trackers = api.content.find(portal_type='Tracker')
            repo_trackers = []
            for brain in trackers:
                tracker = brain.getObject()
                if tracker.repo_url == web_url:
                    repo_trackers.append(tracker)
            if len(repo_trackers) == 0:
                reason = ("No tracker affiliated "
                    "with this GitLab Project: {}").format(web_url)
                logger.error(reason)
                self.request.response.setBody({"error": reason})
                self.request.response.setStatus(500, reason)
                return
            if len(repo_trackers) > 1:
                reason = ("Too many trackers affiliated "
                     "with this GitLab Project: {}".format(len(repo_trackers))
                logger.error(reason)
                self.request.response.setBody({"error": reason})
                self.request.response.setStatus(500)
                return
            tracker = repo_trackers[0]

            # find the poi issue within that tracker
            gitlab_issue = data.get('object_attributes')
            issues = api.content.find(
                    context=tracker,
                    gitlab_issue_id=gitlab_issue.get('iid')
                )
            if len(issues) == 0:
                # this will happen a lot until all GitLab Issues
                # are linked with Poi Issues
                reason = ("There are no Poi Issues linked with this "
                          "GitLab Issue.")
                self.request.response.setBody({"error": reason})
                self.request.response.setStatus(500)
                return
            if len(issues) > 1:
                # this *should* not be possible
                reason = ("There is more than 1 Poi Issue linked with "
                          "this GitLab Issue.")
                self.request.response.setBody({"error": reason})
                self.request.response.setStatus(500)
                return
            issue = issues[0].getObject()

            issue_state = api.content.get_state(issue)
            if issue_state == 'in-progress':
                # if the Poi Issue is already in the desired state,
                # we're done here
                reason = "Poi Issue already in desired state."
                self.request.response.setBody({"done": reason})
                self.request.response.setStatus(200)
                return
            # if one of these labels is on the gitlab issue we need to
            # change the poi issue state to "In Progress"
            state_labels = [
                'dev',
                'In Progress',
                'Needs Review',
                'Needs Work',
                'qa',
                'UAT',
            ]

            changes = data.get('changes')
            if not changes:
                # this should not happen, why would there be a webhook
                # request without changes
                reason = ("No changes sent.")
                self.request.response.setBody({"error": reason})
                self.request.response.setStatus(500)
                return
            changes_labels = changes.get('labels')
            if not changes_labels:
                # if no label changes we're done here
                reason = ("No changes to labels.")
                self.request.response.setBody({"done": reason})
                self.request.response.setStatus(200)
                return
            changes_current_labels = changes_labels.get('current')
            current_labels = []
            for label in changes_current_labels:
                current_labels.append(label.get('title'))
            overlap = list(set(state_labels) & set(current_labels))
            if not overlap:
                # if none of the labels are in the list, we're done here
                reason = ("No changes to labels that matter.")
                self.request.response.setBody({"done": reason})
                self.request.response.setStatus(200)
                return

            try:
                api.content.transition(obj=issue, to_state='in-progress')
            except:
                reason = "Could not transition to desired workflow state."
                self.request.response.setBody({"error": reason})
                self.request.response.setStatus(500)
                return

            reason = "Poi Issue state updated."
            self.request.response.setBody({"done": reason})
            self.request.response.setStatus(200)
            # if we got this far we should add a response which will also
            # trigger notification email
            wftool = getToolByName(issue, 'portal_workflow')
            before_state = wftool.getTitleForStateOnType(issue_state, 'Issue')
            response = Response('State changed via GitLab')
            response.add_change('review_state', 'Issue state',
                                 before_state, 'In Progress')
            folder = IResponseContainer(issue)
            folder.add(response)
            return